"""Tests for the HTTP API Client."""
import pytest

import solana.system_program as sp
from solana.rpc.api import DataSliceOpt
from solana.rpc.async_api import AsyncClient
from solana.keypair import Keypair
from solana.rpc.core import RPCException
from solana.rpc.types import RPCError
from solana.transaction import Transaction
from spl.token.constants import WRAPPED_SOL_MINT

from .utils import AIRDROP_AMOUNT, assert_valid_response


@pytest.mark.integration
@pytest.mark.asyncio
async def test_request_air_drop(async_stubbed_sender: Keypair, test_http_client_async: AsyncClient):
    """Test air drop to async_stubbed_sender."""
    resp = await test_http_client_async.request_airdrop(async_stubbed_sender.public_key, AIRDROP_AMOUNT)
    assert_valid_response(resp)
    await test_http_client_async.confirm_transaction(resp["result"])
    balance = await test_http_client_async.get_balance(async_stubbed_sender.public_key)
    assert balance["result"]["value"] == AIRDROP_AMOUNT


@pytest.mark.integration
@pytest.mark.asyncio
async def test_request_air_drop_prefetched_blockhash(async_stubbed_sender_prefetched_blockhash, test_http_client_async):
    """Test air drop to async_stubbed_sender."""
    resp = await test_http_client_async.request_airdrop(
        async_stubbed_sender_prefetched_blockhash.public_key, AIRDROP_AMOUNT
    )
    assert_valid_response(resp)
    await test_http_client_async.confirm_transaction(resp["result"])
    balance = await test_http_client_async.get_balance(async_stubbed_sender_prefetched_blockhash.public_key)
    assert balance["result"]["value"] == AIRDROP_AMOUNT


@pytest.mark.integration
@pytest.mark.asyncio
async def test_request_air_drop_cached_blockhash(
    async_stubbed_sender_cached_blockhash, test_http_client_async_cached_blockhash
):
    """Test air drop to async_stubbed_sender."""
    resp = await test_http_client_async_cached_blockhash.request_airdrop(
        async_stubbed_sender_cached_blockhash.public_key, AIRDROP_AMOUNT
    )
    assert_valid_response(resp)
    await test_http_client_async_cached_blockhash.confirm_transaction(resp["result"])
    balance = await test_http_client_async_cached_blockhash.get_balance(
        async_stubbed_sender_cached_blockhash.public_key
    )
    assert balance["result"]["value"] == AIRDROP_AMOUNT


@pytest.mark.integration
@pytest.mark.asyncio
async def test_send_invalid_transaction(test_http_client_async):
    """Test sending an invalid transaction to localnet."""
    # Create transfer tx to transfer lamports from stubbed sender to stubbed_receiver
    with pytest.raises(RPCException) as exc_info:
        await test_http_client_async.send_raw_transaction(b"foo")
    assert exc_info.value.args[0].keys() == RPCError.__annotations__.keys()


@pytest.mark.integration
@pytest.mark.asyncio
async def test_send_transaction_and_get_balance(async_stubbed_sender, async_stubbed_receiver, test_http_client_async):
    """Test sending a transaction to localnet."""
    # Create transfer tx to transfer lamports from stubbed sender to async_stubbed_receiver
    transfer_tx = Transaction().add(
        sp.transfer(
            sp.TransferParams(
                from_pubkey=async_stubbed_sender.public_key, to_pubkey=async_stubbed_receiver, lamports=1000
            )
        )
    )
    resp = await test_http_client_async.send_transaction(transfer_tx, async_stubbed_sender)
    assert_valid_response(resp)
    # Confirm transaction
    await test_http_client_async.confirm_transaction(resp["result"])
    # Check balances
    resp = await test_http_client_async.get_balance(async_stubbed_sender.public_key)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 9999994000
    resp = await test_http_client_async.get_balance(async_stubbed_receiver)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 954


@pytest.mark.integration
@pytest.mark.asyncio
async def test_send_transaction_prefetched_blockhash(
    async_stubbed_sender_prefetched_blockhash, async_stubbed_receiver_prefetched_blockhash, test_http_client_async
):
    """Test sending a transaction to localnet."""
    # Create transfer tx to transfer lamports from stubbed sender to async_stubbed_receiver
    transfer_tx = Transaction().add(
        sp.transfer(
            sp.TransferParams(
                from_pubkey=async_stubbed_sender_prefetched_blockhash.public_key,
                to_pubkey=async_stubbed_receiver_prefetched_blockhash,
                lamports=1000,
            )
        )
    )
    resp = await test_http_client_async.send_transaction(transfer_tx, async_stubbed_sender_prefetched_blockhash)
    assert_valid_response(resp)
    # Confirm transaction
    await test_http_client_async.confirm_transaction(resp["result"])
    # Check balances
    resp = await test_http_client_async.get_balance(async_stubbed_sender_prefetched_blockhash.public_key)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 9999994000
    resp = await test_http_client_async.get_balance(async_stubbed_receiver_prefetched_blockhash)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 954


@pytest.mark.integration
@pytest.mark.asyncio
async def test_send_transaction_cached_blockhash(
    async_stubbed_sender_cached_blockhash,
    async_stubbed_receiver_cached_blockhash,
    test_http_client_async_cached_blockhash,
):
    """Test sending a transaction to localnet."""
    # Create transfer tx to transfer lamports from stubbed sender to stubbed_receiver
    transfer_tx = Transaction().add(
        sp.transfer(
            sp.TransferParams(
                from_pubkey=async_stubbed_sender_cached_blockhash.public_key,
                to_pubkey=async_stubbed_receiver_cached_blockhash,
                lamports=1000,
            )
        )
    )
    assert len(test_http_client_async_cached_blockhash.blockhash_cache.unused_blockhashes) == 0
    assert len(test_http_client_async_cached_blockhash.blockhash_cache.used_blockhashes) == 0
    resp = await test_http_client_async_cached_blockhash.send_transaction(
        transfer_tx, async_stubbed_sender_cached_blockhash
    )
    # we could have got a new blockhash or not depending on network latency and luck
    assert len(test_http_client_async_cached_blockhash.blockhash_cache.unused_blockhashes) in (0, 1)
    assert len(test_http_client_async_cached_blockhash.blockhash_cache.used_blockhashes) == 1
    assert_valid_response(resp)
    # Confirm transaction
    await test_http_client_async_cached_blockhash.confirm_transaction(resp["result"])
    # Check balances
    resp = await test_http_client_async_cached_blockhash.get_balance(async_stubbed_sender_cached_blockhash.public_key)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 9999994000

    # Second transaction
    transfer_tx = Transaction().add(
        sp.transfer(
            sp.TransferParams(
                from_pubkey=async_stubbed_sender_cached_blockhash.public_key,
                to_pubkey=async_stubbed_receiver_cached_blockhash,
                lamports=2000,
            )
        )
    )
    resp = await test_http_client_async_cached_blockhash.get_balance(async_stubbed_receiver_cached_blockhash)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 954
    resp = await test_http_client_async_cached_blockhash.send_transaction(
        transfer_tx, async_stubbed_sender_cached_blockhash
    )
    # we could have got a new blockhash or not depending on network latency and luck
    assert len(test_http_client_async_cached_blockhash.blockhash_cache.unused_blockhashes) in (0, 1)
    assert len(test_http_client_async_cached_blockhash.blockhash_cache.used_blockhashes) in (1, 2)
    assert_valid_response(resp)
    # Confirm transaction
    resp = await test_http_client_async_cached_blockhash.confirm_transaction(resp["result"])
    # Check balances
    resp = await test_http_client_async_cached_blockhash.get_balance(async_stubbed_sender_cached_blockhash.public_key)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 9999987000


@pytest.mark.integration
@pytest.mark.asyncio
async def test_send_raw_transaction_and_get_balance(
    async_stubbed_sender, async_stubbed_receiver, test_http_client_async
):
    """Test sending a raw transaction to localnet."""
    # Get a recent blockhash
    resp = await test_http_client_async.get_recent_blockhash()
    assert_valid_response(resp)
    recent_blockhash = resp["result"]["value"]["blockhash"]
    # Create transfer tx transfer lamports from stubbed sender to async_stubbed_receiver
    transfer_tx = Transaction(recent_blockhash=recent_blockhash).add(
        sp.transfer(
            sp.TransferParams(
                from_pubkey=async_stubbed_sender.public_key, to_pubkey=async_stubbed_receiver, lamports=1000
            )
        )
    )
    # Sign transaction
    transfer_tx.sign(async_stubbed_sender)
    # Send raw transaction
    resp = await test_http_client_async.send_raw_transaction(transfer_tx.serialize())
    assert_valid_response(resp)
    # Confirm transaction
    resp = await test_http_client_async.confirm_transaction(resp["result"])
    # Check balances
    resp = await test_http_client_async.get_balance(async_stubbed_sender.public_key)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 9999988000
    resp = await test_http_client_async.get_balance(async_stubbed_receiver)
    assert_valid_response(resp)
    assert resp["result"]["value"] == 1954


@pytest.mark.integration
@pytest.mark.asyncio
async def test_confirm_bad_signature(test_http_client_async: AsyncClient) -> None:
    """Test that RPCException is raised when trying to confirm an invalid signature."""
    with pytest.raises(RPCException) as exc_info:
        await test_http_client_async.confirm_transaction("foo")
    err_object = exc_info.value.args[0]
    assert err_object == {"code": -32602, "message": "Invalid param: WrongSize"}


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_block_commitment(test_http_client_async):
    """Test get block commitment."""
    resp = await test_http_client_async.get_block_commitment(5)
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_block_time(test_http_client_async):
    """Test get block time."""
    resp = await test_http_client_async.get_block_time(5)
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_cluster_nodes(test_http_client_async):
    """Test get cluster nodes."""
    resp = await test_http_client_async.get_cluster_nodes()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_confirmed_block(test_http_client_async):
    """Test get confirmed block."""
    resp = await test_http_client_async.get_confirmed_block(1)
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_confirmed_block_with_encoding(test_http_client_async):
    """Test get confrimed block with encoding."""
    resp = await test_http_client_async.get_confirmed_block(1, encoding="base64")
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_confirmed_blocks(test_http_client_async):
    """Test get confirmed blocks."""
    resp = await test_http_client_async.get_confirmed_blocks(5, 10)
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_confirmed_signature_for_address2(test_http_client_async):
    """Test get confirmed signature for address2."""
    resp = await test_http_client_async.get_confirmed_signature_for_address2(
        "Vote111111111111111111111111111111111111111", limit=1
    )
    assert_valid_response(resp)


# TODO(michael): This RPC call is only available in solana-core v1.7 or newer.
# @pytest.mark.integration
# @pytest.mark.asyncio
# async def test_get_signatures_for_address(test_http_client_async_async):
#     """Test get signatures for addresses."""
#     resp = await test_http_client_async_async.get_signatures_for_address(
#         "Vote111111111111111111111111111111111111111", limit=1
#     )
#     assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_epoch_info(test_http_client_async):
    """Test get epoch info."""
    resp = await test_http_client_async.get_epoch_info()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_epoch_schedule(test_http_client_async):
    """Test get epoch schedule."""
    resp = await test_http_client_async.get_epoch_schedule()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_fee_calculator_for_blockhash(test_http_client_async):
    """Test get fee calculator for blockhash."""
    resp = await test_http_client_async.get_recent_blockhash()
    assert_valid_response(resp)
    resp = await test_http_client_async.get_fee_calculator_for_blockhash(resp["result"]["value"]["blockhash"])
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_slot(test_http_client_async):
    """Test get slot."""
    resp = await test_http_client_async.get_slot()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_fees(test_http_client_async):
    """Test get fees."""
    resp = await test_http_client_async.get_fees()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_first_available_block(test_http_client_async):
    """Test get first available block."""
    resp = await test_http_client_async.get_first_available_block()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_genesis_hash(test_http_client_async):
    """Test get genesis hash."""
    resp = await test_http_client_async.get_genesis_hash()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_identity(test_http_client_async):
    """Test get identity."""
    resp = await test_http_client_async.get_genesis_hash()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_inflation_governor(test_http_client_async):
    """Test get inflation governor."""
    resp = await test_http_client_async.get_inflation_governor()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_inflation_rate(test_http_client_async):
    """Test get inflation rate."""
    resp = await test_http_client_async.get_inflation_rate()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_largest_accounts(test_http_client_async):
    """Test get largest accounts."""
    resp = await test_http_client_async.get_largest_accounts()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_leader_schedule(test_http_client_async):
    """Test get leader schedule."""
    resp = await test_http_client_async.get_leader_schedule()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_minimum_balance_for_rent_exemption(test_http_client_async):
    """Test get minimum balance for rent exemption."""
    resp = await test_http_client_async.get_minimum_balance_for_rent_exemption(50)
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_slot_leader(test_http_client_async):
    """Test get slot leader."""
    resp = await test_http_client_async.get_slot_leader()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_supply(test_http_client_async):
    """Test get slot leader."""
    resp = await test_http_client_async.get_supply()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_token_largest_accounts(test_http_client_async):
    """Test get token largest accounts."""
    resp = await test_http_client_async.get_token_largest_accounts(WRAPPED_SOL_MINT)
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_token_supply(test_http_client_async):
    """Test get token supply."""
    resp = await test_http_client_async.get_token_supply(WRAPPED_SOL_MINT)
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_transaction_count(test_http_client_async):
    """Test get transactinon count."""
    resp = await test_http_client_async.get_transaction_count()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_version(test_http_client_async):
    """Test get version."""
    resp = await test_http_client_async.get_version()
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_account_info(async_stubbed_sender, test_http_client_async):
    """Test get_account_info."""
    resp = await test_http_client_async.get_account_info(async_stubbed_sender.public_key)
    assert_valid_response(resp)
    resp = await test_http_client_async.get_account_info(async_stubbed_sender.public_key, encoding="jsonParsed")
    assert_valid_response(resp)
    resp = await test_http_client_async.get_account_info(async_stubbed_sender.public_key, data_slice=DataSliceOpt(1, 1))
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_multiple_accounts(async_stubbed_sender, test_http_client_async):
    """Test get_multiple_accounts."""
    pubkeys = [async_stubbed_sender.public_key] * 2
    resp = await test_http_client_async.get_multiple_accounts(pubkeys)
    assert_valid_response(resp)
    resp = await test_http_client_async.get_multiple_accounts(pubkeys, encoding="jsonParsed")
    assert_valid_response(resp)
    resp = await test_http_client_async.get_multiple_accounts(pubkeys, data_slice=DataSliceOpt(1, 1))
    assert_valid_response(resp)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_vote_accounts(test_http_client_async):
    """Test get vote accounts."""
    resp = await test_http_client_async.get_vote_accounts()
    assert_valid_response(resp)



"""Integration test utils."""

from solana.rpc.types import RPCResponse

AIRDROP_AMOUNT = 10_000_000_000


def assert_valid_response(resp: RPCResponse):
    """Assert valid RPCResponse."""
    assert resp["jsonrpc"] == "2.0"
    assert resp["id"]
    assert resp["result"]


def compare_responses_without_ids(left: RPCResponse, right: RPCResponse) -> None:
    """Compare RPC responses but ignore IDs."""
    assert {key: val for key, val in left.items() if key != "id"} == {
        key: val for key, val in right.items() if key != "id"
    }



# pylint: disable=R0401
"""Tests for the SPL Token Client."""
import pytest

import spl.token._layouts as layouts
from solana.publickey import PublicKey
from solana.rpc.types import TxOpts
from solana.utils.helpers import decode_byte_string
from spl.token.async_client import AsyncToken
from spl.token.constants import ASSOCIATED_TOKEN_PROGRAM_ID, TOKEN_PROGRAM_ID

from .utils import AIRDROP_AMOUNT, assert_valid_response


@pytest.mark.integration
@pytest.mark.asyncio
@pytest.fixture(scope="module")
async def test_token(async_stubbed_sender, freeze_authority, test_http_client_async) -> AsyncToken:
    """Test create mint."""
    resp = await test_http_client_async.request_airdrop(async_stubbed_sender.public_key, AIRDROP_AMOUNT)
    await test_http_client_async.confirm_transaction(resp["result"])
    assert_valid_response(resp)

    expected_decimals = 6
    token_client = await AsyncToken.create_mint(
        test_http_client_async,
        async_stubbed_sender,
        async_stubbed_sender.public_key,
        expected_decimals,
        TOKEN_PROGRAM_ID,
        freeze_authority.public_key,
    )

    assert token_client.pubkey
    assert token_client.program_id == TOKEN_PROGRAM_ID
    assert token_client.payer.public_key == async_stubbed_sender.public_key

    resp = await test_http_client_async.get_account_info(token_client.pubkey)
    assert_valid_response(resp)
    assert resp["result"]["value"]["owner"] == str(TOKEN_PROGRAM_ID)

    mint_data = layouts.MINT_LAYOUT.parse(decode_byte_string(resp["result"]["value"]["data"][0]))
    assert mint_data.is_initialized
    assert mint_data.decimals == expected_decimals
    assert mint_data.supply == 0
    assert PublicKey(mint_data.mint_authority) == async_stubbed_sender.public_key
    assert PublicKey(mint_data.freeze_authority) == freeze_authority.public_key
    return token_client


@pytest.mark.integration
@pytest.mark.asyncio
@pytest.fixture(scope="module")
async def async_stubbed_sender_token_account_pk(
    async_stubbed_sender, test_token  # pylint: disable=redefined-outer-name
) -> PublicKey:
    """Token account for stubbed sender."""
    return await test_token.create_account(async_stubbed_sender.public_key)


@pytest.mark.integration
@pytest.mark.asyncio
@pytest.fixture(scope="module")
async def async_stubbed_receiver_token_account_pk(
    async_stubbed_receiver, test_token  # pylint: disable=redefined-outer-name
) -> PublicKey:
    """Token account for stubbed receiver."""
    return await test_token.create_account(async_stubbed_receiver)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_new_account(
    async_stubbed_sender, test_http_client_async, test_token
):  # pylint: disable=redefined-outer-name
    """Test creating a new token account."""
    token_account_pk = await test_token.create_account(async_stubbed_sender.public_key)
    resp = await test_http_client_async.get_account_info(token_account_pk)
    assert_valid_response(resp)
    assert resp["result"]["value"]["owner"] == str(TOKEN_PROGRAM_ID)

    account_data = layouts.ACCOUNT_LAYOUT.parse(decode_byte_string(resp["result"]["value"]["data"][0]))
    assert account_data.state
    assert not account_data.amount
    assert (
        not account_data.delegate_option
        and not account_data.delegated_amount
        and PublicKey(account_data.delegate) == PublicKey(0)
    )
    assert not account_data.close_authority_option and PublicKey(account_data.close_authority) == PublicKey(0)
    assert not account_data.is_native_option and not account_data.is_native
    assert PublicKey(account_data.mint) == test_token.pubkey
    assert PublicKey(account_data.owner) == async_stubbed_sender.public_key


@pytest.mark.integration
@pytest.mark.asyncio
async def test_new_associated_account(test_token):  # pylint: disable=redefined-outer-name
    """Test creating a new associated token account."""
    new_acct = PublicKey(0)
    token_account_pubkey = await test_token.create_associated_token_account(new_acct)
    expected_token_account_key, _ = new_acct.find_program_address(
        seeds=[bytes(new_acct), bytes(TOKEN_PROGRAM_ID), bytes(test_token.pubkey)],
        program_id=ASSOCIATED_TOKEN_PROGRAM_ID,
    )
    assert token_account_pubkey == expected_token_account_key


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_account_info(
    async_stubbed_sender, async_stubbed_sender_token_account_pk, test_token
):  # pylint: disable=redefined-outer-name
    """Test get token account info."""
    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.is_initialized is True
    assert account_info.mint == test_token.pubkey
    assert account_info.owner == async_stubbed_sender.public_key
    assert account_info.amount == 0
    assert account_info.delegate is None
    assert account_info.delegated_amount == 0
    assert account_info.is_frozen is False
    assert account_info.is_native is False
    assert account_info.rent_exempt_reserve is None
    assert account_info.close_authority is None


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_mint_info(
    async_stubbed_sender, freeze_authority, test_token
):  # pylint: disable=redefined-outer-name
    """Test get token mint info."""
    mint_info = await test_token.get_mint_info()
    assert mint_info.mint_authority == async_stubbed_sender.public_key
    assert mint_info.supply == 0
    assert mint_info.decimals == 6
    assert mint_info.is_initialized is True
    assert mint_info.freeze_authority == freeze_authority.public_key


@pytest.mark.integration
@pytest.mark.asyncio
async def test_mint_to(
    async_stubbed_sender, async_stubbed_sender_token_account_pk, test_token
):  # pylint: disable=redefined-outer-name
    """Test mint token to account and get balance."""
    expected_amount = 1000
    resp = await test_token.mint_to(
        dest=async_stubbed_sender_token_account_pk,
        mint_authority=async_stubbed_sender,
        amount=1000,
        opts=TxOpts(skip_confirmation=False),
    )
    assert_valid_response(resp)
    resp = await test_token.get_balance(async_stubbed_sender_token_account_pk)
    balance_info = resp["result"]["value"]
    assert balance_info["amount"] == str(expected_amount)
    assert balance_info["decimals"] == 6
    assert balance_info["uiAmount"] == 0.001


@pytest.mark.integration
@pytest.mark.asyncio
async def test_transfer(
    async_stubbed_sender, async_stubbed_receiver_token_account_pk, async_stubbed_sender_token_account_pk, test_token
):  # pylint: disable=redefined-outer-name
    """Test token transfer."""
    expected_amount = 500
    resp = await test_token.transfer(
        source=async_stubbed_sender_token_account_pk,
        dest=async_stubbed_receiver_token_account_pk,
        owner=async_stubbed_sender,
        amount=expected_amount,
        opts=TxOpts(skip_confirmation=False),
    )
    assert_valid_response(resp)
    resp = await test_token.get_balance(async_stubbed_receiver_token_account_pk)
    balance_info = resp["result"]["value"]
    assert balance_info["amount"] == str(expected_amount)
    assert balance_info["decimals"] == 6
    assert balance_info["uiAmount"] == 0.0005


@pytest.mark.integration
@pytest.mark.asyncio
async def test_burn(
    async_stubbed_sender, async_stubbed_sender_token_account_pk, test_token
):  # pylint: disable=redefined-outer-name
    """Test burning tokens."""
    burn_amount = 200
    expected_amount = 300

    burn_resp = await test_token.burn(
        account=async_stubbed_sender_token_account_pk,
        owner=async_stubbed_sender,
        amount=burn_amount,
        multi_signers=None,
        opts=TxOpts(skip_confirmation=False),
    )
    assert_valid_response(burn_resp)

    resp = await test_token.get_balance(async_stubbed_sender_token_account_pk)
    balance_info = resp["result"]["value"]
    assert balance_info["amount"] == str(expected_amount)
    assert balance_info["decimals"] == 6
    assert balance_info["uiAmount"] == 0.0003


@pytest.mark.integration
@pytest.mark.asyncio
async def test_mint_to_checked(
    async_stubbed_sender,
    async_stubbed_sender_token_account_pk,
    test_token,
):  # pylint: disable=redefined-outer-name
    """Test mint token checked and get balance."""
    expected_amount = 1000
    mint_amount = 700
    expected_decimals = 6

    mint_resp = await test_token.mint_to_checked(
        dest=async_stubbed_sender_token_account_pk,
        mint_authority=async_stubbed_sender,
        amount=mint_amount,
        decimals=expected_decimals,
        multi_signers=None,
        opts=TxOpts(skip_confirmation=False),
    )
    assert_valid_response(mint_resp)

    resp = await test_token.get_balance(async_stubbed_sender_token_account_pk)
    balance_info = resp["result"]["value"]
    assert balance_info["amount"] == str(expected_amount)
    assert balance_info["decimals"] == expected_decimals
    assert balance_info["uiAmount"] == 0.001


@pytest.mark.integration
@pytest.mark.asyncio
async def test_transfer_checked(
    async_stubbed_sender, async_stubbed_receiver_token_account_pk, async_stubbed_sender_token_account_pk, test_token
):  # pylint: disable=redefined-outer-name
    """Test token transfer."""
    transfer_amount = 500
    total_amount = 1000
    expected_decimals = 6

    transfer_resp = await test_token.transfer_checked(
        source=async_stubbed_sender_token_account_pk,
        dest=async_stubbed_receiver_token_account_pk,
        owner=async_stubbed_sender,
        amount=transfer_amount,
        decimals=expected_decimals,
        multi_signers=None,
        opts=TxOpts(skip_confirmation=False),
    )
    assert_valid_response(transfer_resp)

    resp = await test_token.get_balance(async_stubbed_receiver_token_account_pk)
    balance_info = resp["result"]["value"]
    assert balance_info["amount"] == str(total_amount)
    assert balance_info["decimals"] == expected_decimals
    assert balance_info["uiAmount"] == 0.001


@pytest.mark.integration
@pytest.mark.asyncio
async def test_burn_checked(
    async_stubbed_sender, async_stubbed_sender_token_account_pk, test_token
):  # pylint: disable=redefined-outer-name
    """Test burning tokens checked."""
    burn_amount = 500
    expected_decimals = 6

    burn_resp = await test_token.burn_checked(
        account=async_stubbed_sender_token_account_pk,
        owner=async_stubbed_sender,
        amount=burn_amount,
        decimals=expected_decimals,
        multi_signers=None,
        opts=TxOpts(skip_confirmation=False),
    )
    assert_valid_response(burn_resp)

    resp = await test_token.get_balance(async_stubbed_sender_token_account_pk)
    balance_info = resp["result"]["value"]
    assert balance_info["amount"] == str(0)
    assert balance_info["decimals"] == expected_decimals
    assert balance_info["uiAmount"] == 0.0


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_accounts(async_stubbed_sender, test_token):  # pylint: disable=redefined-outer-name
    """Test get token accounts."""
    resp = await test_token.get_accounts(async_stubbed_sender.public_key)
    assert_valid_response(resp)
    assert len(resp["result"]["value"]) == 2
    for resp_data in resp["result"]["value"]:
        assert PublicKey(resp_data["pubkey"])
        parsed_data = resp_data["account"]["data"]["parsed"]["info"]
        assert parsed_data["owner"] == str(async_stubbed_sender.public_key)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_approve(
    async_stubbed_sender,
    async_stubbed_receiver,
    async_stubbed_sender_token_account_pk,
    test_token,
    test_http_client_async,
):  # pylint: disable=redefined-outer-name
    """Test approval for delgating a token account."""
    expected_amount_delegated = 500
    resp = await test_token.approve(
        source=async_stubbed_sender_token_account_pk,
        delegate=async_stubbed_receiver,
        owner=async_stubbed_sender.public_key,
        amount=expected_amount_delegated,
    )
    await test_http_client_async.confirm_transaction(resp["result"])
    assert_valid_response(resp)
    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.delegate == async_stubbed_receiver
    assert account_info.delegated_amount == expected_amount_delegated


@pytest.mark.integration
@pytest.mark.asyncio
async def test_revoke(
    async_stubbed_sender,
    async_stubbed_receiver,
    async_stubbed_sender_token_account_pk,
    test_token,
    test_http_client_async,
):  # pylint: disable=redefined-outer-name
    """Test revoke for undelgating a token account."""
    expected_amount_delegated = 500
    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.delegate == async_stubbed_receiver
    assert account_info.delegated_amount == expected_amount_delegated

    revoke_resp = await test_token.revoke(
        account=async_stubbed_sender_token_account_pk,
        owner=async_stubbed_sender.public_key,
    )
    await test_http_client_async.confirm_transaction(revoke_resp["result"])
    assert_valid_response(revoke_resp)
    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.delegate is None
    assert account_info.delegated_amount == 0


@pytest.mark.integration
@pytest.mark.asyncio
async def test_approve_checked(
    async_stubbed_sender,
    async_stubbed_receiver,
    async_stubbed_sender_token_account_pk,
    test_token,
    test_http_client_async,
):  # pylint: disable=redefined-outer-name
    """Test approve_checked for delegating a token account."""
    expected_amount_delegated = 500
    resp = await test_token.approve_checked(
        source=async_stubbed_sender_token_account_pk,
        delegate=async_stubbed_receiver,
        owner=async_stubbed_sender.public_key,
        amount=expected_amount_delegated,
        decimals=6,
    )
    await test_http_client_async.confirm_transaction(resp["result"])
    assert_valid_response(resp)
    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.delegate == async_stubbed_receiver
    assert account_info.delegated_amount == expected_amount_delegated


@pytest.mark.integration
@pytest.mark.asyncio
async def test_freeze_account(
    async_stubbed_sender_token_account_pk, freeze_authority, test_token, test_http_client_async
):  # pylint: disable=redefined-outer-name
    """Test freezing an account."""
    resp = await test_http_client_async.request_airdrop(freeze_authority.public_key, AIRDROP_AMOUNT)
    await test_http_client_async.confirm_transaction(resp["result"])
    assert_valid_response(resp)

    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.is_frozen is False

    freeze_resp = await test_token.freeze_account(async_stubbed_sender_token_account_pk, freeze_authority)
    await test_http_client_async.confirm_transaction(freeze_resp["result"])
    assert_valid_response(freeze_resp)
    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.is_frozen is True


@pytest.mark.integration
@pytest.mark.asyncio
async def test_thaw_account(
    async_stubbed_sender_token_account_pk, freeze_authority, test_token, test_http_client_async
):  # pylint: disable=redefined-outer-name
    """Test thawing an account."""
    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.is_frozen is True

    thaw_resp = await test_token.thaw_account(async_stubbed_sender_token_account_pk, freeze_authority)
    await test_http_client_async.confirm_transaction(thaw_resp["result"])
    assert_valid_response(thaw_resp)
    account_info = await test_token.get_account_info(async_stubbed_sender_token_account_pk)
    assert account_info.is_frozen is False


@pytest.mark.integration
@pytest.mark.asyncio
async def test_close_account(
    async_stubbed_sender,
    async_stubbed_sender_token_account_pk,
    async_stubbed_receiver_token_account_pk,
    test_token,
    test_http_client_async,
):  # pylint: disable=redefined-outer-name
    """Test closing a token account."""
    create_resp = await test_http_client_async.get_account_info(async_stubbed_sender_token_account_pk)
    assert_valid_response(create_resp)
    assert create_resp["result"]["value"]["data"]

    close_resp = await test_token.close_account(
        account=async_stubbed_sender_token_account_pk,
        dest=async_stubbed_receiver_token_account_pk,
        authority=async_stubbed_sender,
    )
    await test_http_client_async.confirm_transaction(close_resp["result"])
    assert_valid_response(close_resp)

    info_resp = await test_http_client_async.get_account_info(async_stubbed_sender_token_account_pk)
    assert_valid_response(info_resp)
    assert info_resp["result"]["value"] is None


@pytest.mark.integration
@pytest.mark.asyncio
async def test_create_multisig(
    async_stubbed_sender, async_stubbed_receiver, test_token, test_http_client
):  # pylint: disable=redefined-outer-name
    """Test creating a multisig account."""
    min_signers = 2
    multisig_pubkey = await test_token.create_multisig(
        min_signers, [async_stubbed_sender.public_key, async_stubbed_receiver]
    )
    resp = test_http_client.get_account_info(multisig_pubkey)
    assert_valid_response(resp)
    assert resp["result"]["value"]["owner"] == str(TOKEN_PROGRAM_ID)

    multisig_data = layouts.MULTISIG_LAYOUT.parse(decode_byte_string(resp["result"]["value"]["data"][0]))
    assert multisig_data.is_initialized
    assert multisig_data.m == min_signers
    assert PublicKey(multisig_data.signer1) == async_stubbed_sender.public_key
    assert PublicKey(multisig_data.signer2) == async_stubbed_receiver
